#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
// std::system includes
#include <iostream>
#include "cycles.h"

// Device code
__global__ void VecAdd( float* A, float* B, float* C, int N ) {
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    if( i < N ) C[ i ] = A[ i ] + B[ i ];
}


// Host code
int main( int argc, char **argv, char **envp ) { 
    ulong N = 32767;
    int CPU = 1, ALL = 1, opt; 
    while ( ( opt = getopt( argc, argv, "n:ca" ) ) != -1 ) {
        switch( opt ) {
            case 'n' :
            {   ulong val;  
                if( sscanf( optarg, "%i", &val ) != 1 ) printf( "illegal value\n" ), exit( EXIT_FAILURE );
                if( val > 0 ) N = val;
                break;
            }
            case 'c' :
                CPU = 0;
                break;
            case 'a' :
                ALL = 0;
                break;
            default :
                printf( "illegal switch\n" ), exit( EXIT_FAILURE );
                break;
       }
    };
    std::cout << ( 0 == CPU ? "CPU" : "GPU" ) << " execution" << std::endl; 

    size_t size = N * sizeof( float );
    float sum = (float)N * ( N + 1. ) / 2.0;

    // Allocate input vectors h_A and h_B in host memory
    float *h_A = (float*)malloc( size ),
          *h_B = (float*)malloc( size ),
          *h_C = (float*)malloc( size ),
          a = 0, b = 0, c = 0;
    if( !h_A || !h_B || !h_C ) perror( "allocation error: " ), exit( EXIT_FAILURE );

    // Initialize input vectors
    for( ulong i = 0; i < N; i++ ) {
       h_A[ i ] = (float)i;
       a += h_A[ i ];
       h_B[ i ] = (float)( N - 1- i );
       b += h_B[ i ];
       h_C[ i ] = 0.;
    }
    std::cout << "A: " << a / sum << ", B: " << b / sum << std::endl;

    // Allocate vectors in device memory
    float *d_A, *d_B, *d_C;
    cudaMalloc( &d_A, size );
    cudaMalloc( &d_B, size );
    cudaMalloc( &d_C, size );

    unsigned calbr = clock_calibrate( 1000000 );
    // Invoke kernel
    int threadsPerBlock = 256,
        blocksPerGrid = ( N -1 + threadsPerBlock ) / threadsPerBlock;
    uint64_t t1, t2;
    t1 = clock_cycles();
    if( 0 != CPU ) {
        // Copy vectors from host memory to device memory
        cudaMemcpy( d_A, h_A, size, cudaMemcpyHostToDevice );
        cudaMemcpy( d_B, h_B, size, cudaMemcpyHostToDevice );
    }
    if( ALL != 0 ) t1 = clock_cycles();

    if( 0 == CPU ) {
        for( ulong i = 0; i < N; i++ )
            h_C[ i ] = h_A[ i ] + h_B[ i ];
    } 
    else {
        VecAdd<<<blocksPerGrid, threadsPerBlock>>>(d_A, d_B, d_C, N);
    }

    t2 = clock_cycles();
    if( 0 != CPU ) {
        // Copy result from device memory to host memory
        cudaMemcpy( h_C, d_C, size, cudaMemcpyDeviceToHost );
    }
    if( 0 == ALL ) t2 = clock_cycles();

    // Free device memory
    cudaFree( d_A );
    cudaFree( d_B );
    cudaFree( d_C );

    for( ulong i = 0; i < N; i++ ) c += h_C[ i ];
    std::cout << "C=A+B: " << c / sum << std::endl;

    std::cout << "time was " << ( t2 - t1 - calbr ) / 1000000 << "."
              << ( t2 - t1 - calbr ) % 1000000 << " million proc. cycles" << std::endl;

    // Free host memory
    //...
    free( h_A );
    free( h_B );
    free( h_C );
}

